import asyncio
import datetime
import functools
import typing
from collections.abc import Callable

import click
import cloup
import dateutil.rrule


class DateType(click.ParamType):
    name = 'iso 8601 date'

    def convert(self, value: typing.Any, param: click.Parameter | None, ctx: click.Context | None) -> typing.Any:  # noqa: ARG002
        if isinstance(value, datetime.date):
            return value
        try:
            return datetime.date.fromisoformat(value)
        except ValueError as e:
            self.fail(str(e))


DATE = DateType()


def to_async(func: Callable) -> Callable:
    @functools.wraps(func)
    def _call(*args, **kwargs):
        asyncio.run(func(*args, **kwargs))

    return _call


def _set_logging(_: click.Context, __: click.Parameter, value: bool):  # noqa: FBT001
    if not value:
        return
    import logging.config

    logging.config.dictConfig(
        {
            'version': 1,
            'handlers': {
                'default': {'class': 'logging.StreamHandler', 'formatter': 'http', 'stream': 'ext://sys.stderr'},
            },
            'formatters': {
                'http': {
                    'format': '[%(asctime)s] %(message)s',
                    'datefmt': '%Y-%m-%d %H:%M:%S',
                },
            },
            'loggers': {
                'httpx': {
                    'handlers': ['default'],
                    'level': 'DEBUG',
                },
            },
        },
    )


generic_options = cloup.option_group(
    'Generic options',
    cloup.option('--less', is_flag=True, help='Print paginated'),
    cloup.option('--http-timeout', type=click.IntRange(min=0), help='Set the http timeout'),
    cloup.option('-v', '--verbose', is_flag=True, callback=_set_logging, is_eager=True, help='Increase verbosity'),
)

upper_lower_bounds = (
    cloup.option('--start', type=DATE, help='Lower included bound date'),
    cloup.option(
        '--end',
        type=DATE,
        default=datetime.datetime.now(tz=datetime.UTC).date(),
        help='Upper included bound date. Defaults to today',
    ),
)


def get_months_in(start: datetime.date, end: datetime.date) -> dateutil.rrule.rrule:
    return dateutil.rrule.rrule(dateutil.rrule.MONTHLY, dtstart=start, until=end)



__version__ = '4.0.0b2'

from factorialhr._api_public import (
    Credentials,
    CredentialsEndpoint,
    WebhookSubscription,
    WebhookSubscriptionEndpoint,
)
from factorialhr._ats import (
    Answer,
    AnswerEndpoint,
    AnswerOriginalQuestionType,
    Application,
    ApplicationEndpoint,
    ApplicationPhase,
    ApplicationPhaseEndpoint,
    ApplicationPhasePhaseType,
    Candidate,
    CandidateEndpoint,
    CandidateGender,
    CandidateSource,
    CandidateSourceCategory,
    CandidateSourceEndpoint,
    EvaluationForm,
    EvaluationFormEndpoint,
    EvaluationFormQuestion,
    Feedback,
    FeedbackEndpoint,
    FeedbackEvaluationFormAnswer,
    HiringStage,
    HiringStageEndpoint,
    HiringStateName,
    JobPosting,
    JobPostingEndpoint,
    JobPostingSalaryPeriod,
    JobPostingStatus,
    Message,
    MessageEndpoint,
    RejectReason,
    RejectReasonDecisionMaker,
    RejectReasonEndpoint,
)
from factorialhr._attendance import (
    BreakConfiguration,
    BreakConfigurationEndpoint,
    EditTimesheetRequest,
    EditTimesheetRequestEndpoint,
    EditTimesheetRequestRequestType,
    EstimatedTime,
    EstimatedTimeEndpoint,
    EstimatedTimeSource,
    OpenShift,
    OpenShiftEndpoint,
    OpenShiftStatus,
    OverTimeRequest,
    OverTimeRequestEndpoint,
    OverTimeRequestStatus,
    Shift,
    ShiftEndpoint,
    ShiftHalfDay,
    WorkedTime,
    WorkedTimeDayType,
    WorkedTimeEndpoint,
)
from factorialhr._client import AccessTokenAuth, ApiClient, ApiKeyAuth, RefreshTokenAuth, RefreshTokenAuthFile
from factorialhr._common import LocationType, Meta, TimeUnit
from factorialhr._employees import (
    Employee,
    EmployeeBankNumberFormat,
    EmployeeEndpoint,
)
from factorialhr._project_management import (
    ExpenseRecord,
    ExpenseRecordEndpoint,
    ExportableExpense,
    ExportableExpenseEndpoint,
    ExportableProject,
    ExportableProjectEndpoint,
    FlexibleTimeRecord,
    FlexibleTimeRecordComment,
    FlexibleTimeRecordCommentEndpoint,
    FlexibleTimeRecordEndpoint,
    Project,
    ProjectEmployeeAssignment,
    ProjectEndpoint,
    ProjectStatus,
    ProjectTask,
    ProjectTaskEndpoint,
    ProjectWorker,
    ProjectWorkerEndpoint,
    Subproject,
    SubprojectEndpoint,
    TimeRecord,
    TimeRecordEndpoint,
)
from factorialhr._teams import (
    Membership,
    MembershipEndpoint,
    Team,
    TeamEndpoint,
)
from factorialhr._time_planning import (
    PlanningVersion,
    PlanningVersionEndpoint,
)

__all__ = [
    'AccessTokenAuth',
    'Answer',
    'AnswerEndpoint',
    'AnswerOriginalQuestionType',
    'ApiClient',
    'ApiKeyAuth',
    'Application',
    'ApplicationEndpoint',
    'ApplicationPhase',
    'ApplicationPhaseEndpoint',
    'ApplicationPhasePhaseType',
    'BreakConfiguration',
    'BreakConfigurationEndpoint',
    'Candidate',
    'CandidateEndpoint',
    'CandidateGender',
    'CandidateSource',
    'CandidateSourceCategory',
    'CandidateSourceEndpoint',
    'Credentials',
    'CredentialsEndpoint',
    'EditTimesheetRequest',
    'EditTimesheetRequestEndpoint',
    'EditTimesheetRequestRequestType',
    'Employee',
    'EmployeeBankNumberFormat',
    'EmployeeEndpoint',
    'EstimatedTime',
    'EstimatedTimeEndpoint',
    'EstimatedTimeSource',
    'EvaluationForm',
    'EvaluationFormEndpoint',
    'EvaluationFormQuestion',
    'ExpenseRecord',
    'ExpenseRecordEndpoint',
    'ExportableExpense',
    'ExportableExpenseEndpoint',
    'ExportableProject',
    'ExportableProjectEndpoint',
    'Feedback',
    'FeedbackEndpoint',
    'FeedbackEvaluationFormAnswer',
    'FlexibleTimeRecord',
    'FlexibleTimeRecordComment',
    'FlexibleTimeRecordCommentEndpoint',
    'FlexibleTimeRecordEndpoint',
    'HiringStage',
    'HiringStageEndpoint',
    'HiringStateName',
    'JobPosting',
    'JobPostingEndpoint',
    'JobPostingSalaryPeriod',
    'JobPostingStatus',
    'LocationType',
    'LocationType',
    'Membership',
    'MembershipEndpoint',
    'Message',
    'MessageEndpoint',
    'Meta',
    'Meta',
    'OpenShift',
    'OpenShiftEndpoint',
    'OpenShiftStatus',
    'OverTimeRequest',
    'OverTimeRequestEndpoint',
    'OverTimeRequestStatus',
    'PlanningVersion',
    'PlanningVersionEndpoint',
    'Project',
    'ProjectEmployeeAssignment',
    'ProjectEndpoint',
    'ProjectStatus',
    'ProjectTask',
    'ProjectTaskEndpoint',
    'ProjectWorker',
    'ProjectWorkerEndpoint',
    'RefreshTokenAuth',
    'RefreshTokenAuthFile',
    'RejectReason',
    'RejectReasonDecisionMaker',
    'RejectReasonEndpoint',
    'Shift',
    'ShiftEndpoint',
    'ShiftHalfDay',
    'Subproject',
    'SubprojectEndpoint',
    'Team',
    'TeamEndpoint',
    'TimeRecord',
    'TimeRecordEndpoint',
    'TimeUnit',
    'TimeUnit',
    'WebhookSubscription',
    'WebhookSubscriptionEndpoint',
    'WorkedTime',
    'WorkedTimeDayType',
    'WorkedTimeEndpoint',
]



import typing

import pydantic

from factorialhr import _common
from factorialhr._client import Endpoint


class Membership(pydantic.BaseModel):
    id: int
    company_id: int | None
    employee_id: int
    team_id: int
    lead: bool


class _MembershipRoot(pydantic.RootModel):
    root: list[Membership]


class MembershipEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/teams/memberships'

    async def all(
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        lead: bool | None = None,
        team_ids: typing.Sequence[int] | None = None,
        employee_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> list[Membership]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-teams-memberships."""
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'lead': lead,
                'team_ids[]': team_ids,
                'employee_ids[]': employee_ids,
            },
        )
        return _MembershipRoot.model_validate(
            await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    @typing.overload
    async def get(self, *, membership_id: int, **kwargs) -> Membership: ...

    @typing.overload
    async def get(
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        lead: bool | None = None,
        team_ids: typing.Sequence[int] | None = None,
        employee_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> tuple[list[Membership], _common.Meta]: ...
    async def get(
        self,
        *,
        membership_id: int | None = None,
        ids: typing.Sequence[int] | None = None,
        lead: bool | None = None,
        team_ids: typing.Sequence[int] | None = None,
        employee_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-teams-memberships-id."""
        if membership_id is not None:
            return Membership.model_validate(await self.api.get(self.endpoint, membership_id, **kwargs))
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'lead': lead,
                'team_ids[]': team_ids,
                'employee_ids[]': employee_ids,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _MembershipRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class Team(pydantic.BaseModel):
    id: int
    name: str
    description: str | None
    avatar: str | None
    employee_ids: list[int] | None
    lead_ids: list[int] | None
    company_id: int


class _TeamRoot(pydantic.RootModel):
    root: list[Team]


class TeamEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/teams/teams'

    async def all(self, *, ids: typing.Sequence[int] | None = None, **kwargs) -> list[Team]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-teams-teams."""
        params = kwargs.get('params', {})
        params.update({'ids[]': ids})
        return _TeamRoot.model_validate(
            await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    @typing.overload
    async def get(self, *, team_id: int, **kwargs) -> Team: ...

    @typing.overload
    async def get(self, *, ids: typing.Sequence[int] | None = None, **kwargs) -> tuple[list[Team], _common.Meta]: ...

    async def get(self, *, team_id: int | None = None, ids: typing.Sequence[int] | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-teams-teams-id."""
        if team_id is not None:
            return Team.model_validate(await self.api.get(self.endpoint, team_id, **kwargs))
        params = kwargs.get('params', {})
        params.update({'ids[]': ids})
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _TeamRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])



import datetime
import enum
import typing

import pydantic

from factorialhr import _common
from factorialhr._client import Endpoint


class EmployeeBankNumberFormat(enum.StrEnum):
    iban = 'iban'
    sort_code_and_account_number = 'sort_code_and_account_number'
    routing_number_and_account_number = 'routing_number_and_account_number'
    clabe = 'clabe'
    other = 'other'
    bank_name_and_account_number = 'bank_name_and_account_number'


class Employee(pydantic.BaseModel):
    id: int
    access_id: int
    first_name: str
    last_name: str
    full_name: str
    preferred_name: str | None
    birth_name: str | None
    gender: str | None
    identifier: str | None
    identifier_type: str | None
    email: str | None
    login_email: str | None
    birthday_on: datetime.date | None
    nationality: str | None
    address_line_1: str | None
    address_line_2: str | None
    postal_code: str | None  #  should be int but api returns string
    city: str | None
    state: str | None
    country: str | None
    bank_number: str | None
    swift_bic: str | None
    bank_number_format: str | None
    company_id: int
    legal_entity_id: int | None
    location_id: int
    created_at: datetime.datetime
    updated_at: datetime.datetime
    social_security_number: int | None
    is_terminating: bool
    terminated_on: datetime.date | None
    termination_reason: str | None
    termination_observations: str | None
    manager_id: int | None
    timeoff_manager_id: int | None
    phone_number: str | None  # should be int but api returns string
    company_identifier: str | None
    age_number: int | None
    contact_name: str | None
    contact_number: str | None  # should be int but api returns string
    personal_email: str | None
    pronouns: str | None
    active: bool | None
    disability_percentage_cents: int | None
    identifier_expiration_date: str | None


class _EmployeeRoot(pydantic.RootModel):
    root: list[Employee]


class EmployeeEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/employees/employees'

    async def all(  # noqa: PLR0913
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        access_ids: typing.Sequence[int] | None = None,
        emails: typing.Sequence[str] | None = None,
        full_text_name: str | None = None,
        legal_entity_ids: typing.Sequence[int] | None = None,
        only_active: bool | None = None,
        team_ids: typing.Sequence[int] | None = None,
        location_ids: typing.Sequence[int] | None = None,
        only_managers: bool | None = None,
        name_starts_with: str | None = None,
        **kwargs,
    ) -> list[Employee]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-employees-employees."""
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'access_ids[]': access_ids,
                'emails[]': emails,
                'full_text_name': full_text_name,
                'legal_entity_ids[]': legal_entity_ids,
                'only_active': only_active,
                'team_ids[]': team_ids,
                'location_ids[]': location_ids,
                'only_managers': only_managers,
                'name_starts_with': name_starts_with,
            },
        )
        return _EmployeeRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    @typing.overload
    async def get(self, *, employee_id: int, **kwargs) -> Employee: ...
    @typing.overload
    async def get(
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        access_ids: typing.Sequence[int] | None = None,
        emails: typing.Sequence[str] | None = None,
        full_text_name: str | None = None,
        legal_entity_ids: typing.Sequence[int] | None = None,
        only_active: bool | None = None,
        team_ids: typing.Sequence[int] | None = None,
        location_ids: typing.Sequence[int] | None = None,
        only_managers: bool | None = None,
        name_starts_with: str | None = None,
        **kwargs,
    ) -> tuple[list[Employee], _common.Meta]: ...

    async def get(  # noqa: PLR0913
        self,
        *,
        employee_id: int | None = None,
        ids: typing.Sequence[int] | None = None,
        access_ids: typing.Sequence[int] | None = None,
        emails: typing.Sequence[str] | None = None,
        full_text_name: str | None = None,
        legal_entity_ids: typing.Sequence[int] | None = None,
        only_active: bool | None = None,
        team_ids: typing.Sequence[int] | None = None,
        location_ids: typing.Sequence[int] | None = None,
        only_managers: bool | None = None,
        name_starts_with: str | None = None,
        **kwargs,
    ):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-employees-employees-id."""
        if employee_id is not None:
            return Employee.model_validate(await self.api.get(self.endpoint, employee_id, **kwargs))
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'access_ids[]': access_ids,
                'emails[]': emails,
                'full_text_name': full_text_name,
                'legal_entity_ids[]': legal_entity_ids,
                'only_active': only_active,
                'team_ids[]': team_ids,
                'location_ids[]': location_ids,
                'only_managers': only_managers,
                'name_starts_with': name_starts_with,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _EmployeeRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])



from factorialhr._cli import account, projects

try:
    import cloup
except ImportError as import_error:
    raise ImportError('Cli not installed. Please install using "pip install factorialhr[cli]".') from import_error


@cloup.group()
def entrypoint():
    pass


entrypoint.add_command(account.account)
entrypoint.add_command(projects.project_entrypoint)



import asyncio
import ssl
import urllib.parse
import uuid
import webbrowser

import aiohttp
from aiohttp import web

from factorialhr._cli import http_server
from factorialhr._client import AccessTokenResponse

HTTP_SERVER = 'localhost'
HTTP_SERVER_PORT = 50101


async def get_authorization_code(
    target_url: str,
    client_id: str,
    redirect_uri: str,
    scope: str,
    ssl_context: ssl.SSLContext | None,
) -> str | None:
    state_secret = str(uuid.uuid4())
    code_queue: asyncio.Queue[str | None] = asyncio.Queue(1)

    async def _handler(request: web.BaseRequest) -> web.StreamResponse:
        if not (state := request.query.get('state', '')):
            await code_queue.put(None)
            return web.Response(status=400, text='state parameter is missing')
        if state != state_secret:
            await code_queue.put(None)
            return web.Response(status=401, text='state parameter is invalid')
        if not (code := request.query.get('code', '')):
            await code_queue.put(None)
            return web.Response(status=400, text='code parameter is missing')
        await code_queue.put(code)
        return web.Response(text='Authorization code received. You can close this window now')

    async with http_server.HttpServer(HTTP_SERVER, HTTP_SERVER_PORT, _handler, ssl_context):
        url = (
            f'{target_url}/oauth/authorize?'
            f'client_id={client_id}&'
            f'redirect_uri={urllib.parse.quote(redirect_uri)}&'
            'response_type=code&'
            f'scope={scope}&'
            f'state={state_secret}'
        )
        print('Open the url in a browser and log in', url)
        webbrowser.open(url)
        return await code_queue.get()


async def get_access_token(  # noqa: PLR0913
    target_url: str,
    client_id: str,
    client_secret: str,
    redirect_uri: str,
    scope: str,
    http_server_timeout: float | None,
    ssl_context: ssl.SSLContext | None,
) -> AccessTokenResponse:
    """Obtain an access token."""
    async with asyncio.timeout(http_server_timeout):
        authorization_code = await get_authorization_code(
            target_url=target_url,
            client_id=client_id,
            redirect_uri=redirect_uri,
            scope=scope,
            ssl_context=ssl_context,
        )
    if authorization_code is None:
        raise RuntimeError('Authorization code could not be obtained')
    async with aiohttp.ClientSession() as session:
        response = await session.post(
            f'{target_url}/oauth/token',
            data={
                'client_id': client_id,
                'client_secret': client_secret,
                'code': authorization_code,
                'grant_type': 'authorization_code',
                'redirect_uri': redirect_uri,
            },
        )
    response.raise_for_status()
    return AccessTokenResponse.model_validate(await response.json())



import asyncio
import pathlib

import click
import cloup
import httpx

import factorialhr
from factorialhr._cli import access_token
from factorialhr._cli import common as common_cli
from factorialhr._client import ApiClient, RefreshTokenAuth, RefreshTokenAuthFile


async def _print_account_info(auth: RefreshTokenAuth, url: str):
    async with ApiClient(auth=auth, base_url=url) as api:
        credentials, _ = await factorialhr.CredentialsEndpoint(api).get()
    if len(credentials) > 1:
        raise ValueError('More than one credentials is currently not supported')
    click.echo(
        f'Logged in as {credentials[0].full_name} on {url}. Session expires at {auth.refresh_token_expiration} '
        f'{auth.created_at.tzname()}.',
    )


def get_session() -> tuple[RefreshTokenAuthFile, str]:
    if not SESSION_FILE.exists():
        raise click.ClickException('No session found. Login with "factorialhr account login".')
    auth, url = RefreshTokenAuthFile.from_file(SESSION_FILE)
    if auth.is_refresh_token_expired():
        msg = f'Session expired at {auth.refresh_token_expiration}. Login with "factorialhr account login".'
        raise click.ClickException(msg)
    return auth, url


async def _logout():
    try:
        auth, url = get_session()
    except click.ClickException:
        SESSION_FILE.unlink(missing_ok=True)
        return

    async with httpx.AsyncClient(base_url=url) as client:
        await asyncio.gather(
            client.post(
                'oauth/revoke',
                data={'client_id': auth.client_id, 'client_secret': auth.client_secret, 'token': auth.access_token},
            ),
            client.post(
                'oauth/revoke',
                data={'client_id': auth.client_id, 'client_secret': auth.client_secret, 'token': auth.refresh_token},
            ),
        )
    SESSION_FILE.unlink()


async def _login(client_id: str, client_secret: str, url: str, scope: str, redirect_url: str):
    try:
        response = await access_token.get_access_token(
            client_id=client_id,
            client_secret=client_secret,
            target_url=url,
            redirect_uri=redirect_url,
            scope=scope,
            http_server_timeout=None,
            ssl_context=None,
        )
        print(response)
    except asyncio.CancelledError:
        raise click.ClickException('Operation cancelled')  # noqa: B904
    except RuntimeError as e:
        raise click.ClickException(str(e)) from e
    auth = RefreshTokenAuthFile(
        file=SESSION_FILE,
        refresh_token=response.refresh_token,
        client_id=client_id,
        client_secret=client_secret,
        access_token=response.access_token,
        token_type=response.token_type,
        created_at=response.created_at,
    )
    auth.to_file(url)
    await _print_account_info(auth, url)


@cloup.group(help='Manages an oauth2 session.')
def account():
    pass


APP_DIR = pathlib.Path(click.get_app_dir('factorialhr_api'))
APP_DIR.mkdir(exist_ok=True)
SESSION_FILE = APP_DIR.joinpath('session.json')


@cloup.command(help='Create an oauth2 session.')
@cloup.argument('client_id', type=str)
@cloup.argument('client_secret', type=str)
@cloup.option('--demo', help='Whether to use the production or demo api.', is_flag=True)
@cloup.option('--scope', help='Oauth2 access scope. Defaults to "read".', default='read')
@cloup.option(
    '--redirect-url',
    help='Where to redirect for oauth2 callback',
    default=f'https://redirectmeto.com/http://{access_token.HTTP_SERVER}:{access_token.HTTP_SERVER_PORT}/authorize/authorization_code',
)
@common_cli.to_async
async def login(client_id: str, client_secret: str, demo: bool, scope: str, redirect_url: str):  # noqa: FBT001
    url = 'https://api.factorialhr.com' if not demo else 'https://api.demo.factorial.dev'
    await _login(client_id=client_id, client_secret=client_secret, url=url, scope=scope, redirect_url=redirect_url)


@cloup.command(help='Get information about the current session.')
def status():
    auth, _ = get_session()
    click.echo(
        f'Session expires at {auth.refresh_token_expiration} {auth.created_at.tzname()}.',
    )


@cloup.command(help='Refresh an already created session.')
@common_cli.to_async
async def refresh():
    auth, url = get_session()
    await auth.refresh_access_token(url)
    await _print_account_info(auth, url)


@cloup.command(help='Invalidate the session.')
@common_cli.to_async
async def logout():
    await _logout()


account.add_command(login)
account.add_command(status)
account.add_command(refresh)
account.add_command(logout)



import typing

import pydantic

from factorialhr import _common
from factorialhr._client import Endpoint


class Credentials(pydantic.BaseModel):
    company_id: int
    id: str
    email: str | None
    login_email: str | None
    full_name: str | None
    first_name: str | None
    last_name: str | None
    employee_id: int | None  # can be None if request is made using an api key
    role: str | None
    gdpr_tos: bool | None
    legal_name: str | None
    locale: str | None
    logo: str | None
    name: str | None
    onboarded_on: str | None
    subscription_plan: str | None
    tin: str | None
    to_be_deleted: str | None
    tos: bool | None


class _CredentialsRoot(pydantic.RootModel):
    root: list[Credentials]


class CredentialsEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/api_public/credentials'

    async def all(self, **kwargs) -> list[Credentials]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-api-public-credentials."""
        return _CredentialsRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    async def get(self, **kwargs) -> tuple[list[Credentials], _common.Meta]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-api-public-credentials."""
        result = await self.api.get(self.endpoint, **kwargs)
        return _CredentialsRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class WebhookSubscription(pydantic.BaseModel):
    id: int
    target_url: str
    type: str
    company_id: int | None
    name: str | None
    challenge: str | None
    enabled: bool
    api_version: str


class _WebhookSubscriptionRoot(pydantic.RootModel):
    root: list[WebhookSubscription]


class WebhookSubscriptionEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/api_public/webhook_subscriptions'

    async def all(self, **kwargs) -> list[WebhookSubscription]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-api-public-webhook-subscriptions."""
        return _WebhookSubscriptionRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, webhook_subscription_id: int, **kwargs) -> WebhookSubscription: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[WebhookSubscription], _common.Meta]: ...

    async def get(self, *, webhook_subscription_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-api-public-webhook-subscriptions-id."""
        result = await self.api.get(self.endpoint, webhook_subscription_id, **kwargs)
        return (
            WebhookSubscription.model_validate(result)
            if webhook_subscription_id is not None
            else (
                _WebhookSubscriptionRoot.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )



import datetime
import typing

import pydantic

from factorialhr import _common
from factorialhr._client import Endpoint


class PlanningVersion(pydantic.BaseModel):
    id: int | None
    effective_at: datetime.date
    planning_tool: str
    number_of_rest_days_in_cents: int | None
    employee_id: int
    work_schedule_schedule_id: int | None


class _PlanningVersionRoot(pydantic.RootModel):
    root: list[PlanningVersion]


class PlanningVersionEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/time_planning/planning_versions'

    async def all(
        self,
        *,
        employee_ids: typing.Sequence[int] | None = None,
        for_shifts: bool | None = None,
        only_active: bool | None = None,
        planning_tool: str | None = None,
        schedule_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> list[PlanningVersion]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-time-planning-planning-versions."""
        params = kwargs.get('params', {})
        params.update(
            {
                'employee_ids[]': employee_ids,
                'for_shifts': for_shifts,
                'only_active': only_active,
                'planning_tool': planning_tool,
                'schedule_ids[]': schedule_ids,
            },
        )
        return _PlanningVersionRoot.model_validate(
            await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    async def get(
        self,
        *,
        employee_ids: typing.Sequence[int] | None = None,
        for_shifts: bool | None = None,
        only_active: bool | None = None,
        planning_tool: str | None = None,
        schedule_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> tuple[list[PlanningVersion], _common.Meta]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-time-planning-planning-versions."""
        params = kwargs.get('params', {})
        params.update(
            {
                'employee_ids[]': employee_ids,
                'for_shifts': for_shifts,
                'only_active': only_active,
                'planning_tool': planning_tool,
                'schedule_ids[]': schedule_ids,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _PlanningVersionRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])



import datetime
import enum
import typing

import pydantic

from factorialhr import _common
from factorialhr._client import Endpoint


class ExpenseRecord(pydantic.BaseModel):
    id: int
    project_worker_id: int
    expense_id: int
    subproject_id: int | None
    original_amount_currency: str | None
    original_amount_cents: int | None
    legal_entity_amount_currency: str | None
    legal_entity_amount_cents: str | None
    effective_on: datetime.date | None
    exchange_rate: int | None
    status: str | None


class _ExpenseRecordRoot(pydantic.RootModel):
    root: list[ExpenseRecord]


class ExpenseRecordEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/project_management/expense_records'

    async def all(  # noqa: PLR0913
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        start_date: datetime.date | None = None,
        end_date: datetime.date | None = None,
        expenses_ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        updated_after: datetime.date | None = None,
        employee_user_name_like: str | None = None,
        project_worker_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> list[ExpenseRecord]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-expense-records."""
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'start_date': start_date,
                'end_date': end_date,
                'expenses_ids[]': expenses_ids,
                'project_ids[]': project_ids,
                'subproject_ids[]': subproject_ids,
                'updated_after': updated_after,
                'employee_user_name_like': employee_user_name_like,
                'project_worker_ids[]': project_worker_ids,
            },
        )
        return _ExpenseRecordRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    @typing.overload
    async def get(self, *, expense_record_id: int, **kwargs) -> ExpenseRecord: ...

    @typing.overload
    async def get(
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        start_date: datetime.date | None = None,
        end_date: datetime.date | None = None,
        expenses_ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        updated_after: datetime.date | None = None,
        employee_user_name_like: str | None = None,
        project_worker_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> tuple[list[ExpenseRecord], _common.Meta]: ...

    async def get(  # noqa: PLR0913
        self,
        *,
        expense_record_id: int | None = None,
        ids: typing.Sequence[int] | None = None,
        start_date: datetime.date | None = None,
        end_date: datetime.date | None = None,
        expenses_ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        updated_after: datetime.date | None = None,
        employee_user_name_like: str | None = None,
        project_worker_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-expense-records-id."""
        if expense_record_id is not None:
            return ExpenseRecord.model_validate(await self.api.get(self.endpoint, expense_record_id, **kwargs))
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'start_date': start_date,
                'end_date': end_date,
                'expenses_ids[]': expenses_ids,
                'project_ids[]': project_ids,
                'subproject_ids[]': subproject_ids,
                'updated_after': updated_after,
                'employee_user_name_like': employee_user_name_like,
                'project_worker_ids[]': project_worker_ids,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _ExpenseRecordRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class ExportableExpense(pydantic.BaseModel):
    date: datetime.date | None
    project_name: str | None
    subproject_name: str | None
    employee_name: str
    preferred_name: str | None
    amount: str | None
    currency: str | None
    expense_category: str | None
    expense_subcategory: str | None
    expense_status: str | None
    expense_link: str | None


class _ExportableExpenseRoot(pydantic.RootModel):
    root: list[ExportableExpense]


class ExportableExpenseEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/project_management/exportable_expenses'

    async def all(
        self,
        *,
        start_date: datetime.date | None = None,
        end_date: datetime.date | None = None,
        project_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> list[ExportableExpense]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-exportable-expenses."""
        params = kwargs.get('params', {})
        params.update(
            {
                'start_date': start_date,
                'end_date': end_date,
                'project_ids[]': project_ids,
            },
        )
        return _ExportableExpenseRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    async def get(
        self,
        *,
        start_date: datetime.date | None = None,
        end_date: datetime.date | None = None,
        project_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> tuple[list[ExportableExpense], _common.Meta]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-exportable-expenses."""
        params = kwargs.get('params', {})
        params.update(
            {
                'start_date': start_date,
                'end_date': end_date,
                'project_ids[]': project_ids,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _ExportableExpenseRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class ExportableProject(pydantic.BaseModel):
    id: str
    date: datetime.date | None
    project_name: str
    project_code: str | None
    project_start_date: datetime.date | None
    project_due_date: datetime.date | None
    project_status: str
    subproject_name: str | None
    employee_name: str | None
    employee_id: int | None
    inputed_time: float  # api returns string but can actually be a float


class _ExportableProjectRoot(pydantic.RootModel):
    root: list[ExportableProject]


class ExportableProjectEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/project_management/exportable_projects'

    async def all(
        self,
        *,
        start_date: datetime.date | None = None,
        end_date: datetime.date | None = None,
        project_ids: typing.Sequence[int] | None = None,
        time_format: str | None = None,
        include_date: bool | None = None,
        **kwargs,
    ) -> list[ExportableProject]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-exportable-projects."""
        params = kwargs.get('params', {})
        params.update(
            {
                'start_date': start_date,
                'end_date': end_date,
                'project_ids[]': project_ids,
                'time_format': time_format,
                'include_date': include_date,
            },
        )
        return _ExportableProjectRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    async def get(
        self,
        *,
        start_date: datetime.date | None = None,
        end_date: datetime.date | None = None,
        project_ids: typing.Sequence[int] | None = None,
        time_format: str | None = None,
        include_date: bool | None = None,
        **kwargs,
    ) -> tuple[list[ExportableProject], _common.Meta]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-exportable-projects."""
        params = kwargs.get('params', {})
        params.update(
            {
                'start_date': start_date,
                'end_date': end_date,
                'project_ids[]': project_ids,
                'time_format': time_format,
                'include_date': include_date,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _ExportableProjectRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class FlexibleTimeRecord(pydantic.BaseModel):
    id: int
    date: datetime.date
    imputed_minutes: int
    project_worker_id: int
    subproject_id: int | None


class _FlexibleTimeRecordRoot(pydantic.RootModel):
    root: list[FlexibleTimeRecord]


class FlexibleTimeRecordEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/project_management/flexible_time_records'

    async def all(
        self,
        *,
        ids: typing.Sequence[int] | None = None,  # documentation says it's required but it's not
        project_worker_ids: typing.Sequence[int] | None = None,  # documentation says it's required but it's not
        starts_on: datetime.date | None = None,
        ends_on: datetime.date | None = None,
        updated_after: str | None = None,  # last time I checked it seems to have no effect
        **kwargs,
    ) -> list[FlexibleTimeRecord]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-flexible-time-records."""
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'project_worker_ids[]': project_worker_ids,
                'starts_on': starts_on,
                'ends_on': ends_on,
                'updated_after': updated_after,
            },
        )
        return _FlexibleTimeRecordRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    @typing.overload
    async def get(self, *, flexible_time_record_id: int, **kwargs) -> FlexibleTimeRecord: ...

    @typing.overload
    async def get(
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        project_worker_ids: typing.Sequence[int] | None = None,
        starts_on: datetime.date | None = None,
        ends_on: datetime.date | None = None,
        updated_after: str | None = None,
        **kwargs,
    ) -> tuple[list[FlexibleTimeRecord], _common.Meta]: ...

    async def get(  # noqa: PLR0913
        self,
        *,
        flexible_time_record_id: int | None = None,
        ids: typing.Sequence[int] | None = None,
        project_worker_ids: typing.Sequence[int] | None = None,
        starts_on: datetime.date | None = None,
        ends_on: datetime.date | None = None,
        updated_after: str | None = None,
        **kwargs,
    ):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-flexible-time-records-id."""
        if flexible_time_record_id is not None:
            return FlexibleTimeRecord.model_validate(
                await self.api.get(self.endpoint, flexible_time_record_id, **kwargs),
            )

        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'project_worker_ids[]': project_worker_ids,
                'starts_on': starts_on,
                'ends_on': ends_on,
                'updated_after': updated_after,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return (
            _FlexibleTimeRecordRoot.model_validate(result['data']).root,
            _common.Meta.model_validate(result['meta']),
        )


class FlexibleTimeRecordComment(pydantic.BaseModel):
    id: int
    content: str
    flexible_time_record_id: int


class _FlexibleTimeRecordCommentRoot(pydantic.RootModel):
    root: list[FlexibleTimeRecordComment]


class FlexibleTimeRecordCommentEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/project_management/flexible_time_record_comments'

    async def all(self, **kwargs) -> list[FlexibleTimeRecordComment]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-flexible-time-record-comments."""
        return _FlexibleTimeRecordCommentRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, flexible_time_record_comment_id: int, **kwargs) -> FlexibleTimeRecordComment: ...

    @typing.overload
    async def get(
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        flexible_time_record_id: int | None = None,
        **kwargs,
    ) -> tuple[list[FlexibleTimeRecordComment], _common.Meta]: ...

    async def get(
        self,
        *,
        flexible_time_record_comment_id: int | None = None,
        ids: typing.Sequence[int] | None = None,
        flexible_time_record_id: int | None = None,
        **kwargs,
    ):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-flexible-time-record-comments-id."""
        if flexible_time_record_comment_id is not None:
            return FlexibleTimeRecordComment.model_validate(
                await self.api.get(self.endpoint, flexible_time_record_comment_id, **kwargs),
            )

        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'flexible_time_record_id': flexible_time_record_id,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _FlexibleTimeRecordCommentRoot.model_validate(result['data']).root, _common.Meta.model_validate(
            result['meta'],
        )


class ProjectStatus(enum.StrEnum):
    active = 'active'
    closed = 'closed'
    processing = 'processing'


class ProjectEmployeeAssignment(enum.StrEnum):
    manual = 'manual'
    company = 'company'


class Project(pydantic.BaseModel):
    id: int
    name: str
    code: str | None
    start_date: datetime.date | None
    due_date: datetime.date | None
    status: ProjectStatus
    employees_assignment: ProjectEmployeeAssignment
    inputed_minutes: int | None
    is_billable: bool
    fixed_cost_cents: int | None
    labor_cost_cents: int | None
    legal_entity_id: int
    spending_cost_cents: int | None
    client_id: int | None
    total_cost_cents: int | None


class _ProjectRoot(pydantic.RootModel):
    root: list[Project]


class ProjectEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/project_management/projects'

    async def all(  # noqa: PLR0913
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        name: str | None = None,
        name_or_code: str | None = None,
        include_inputed_minutes: bool | None = None,
        updated_after: datetime.date | None = None,
        legal_entity_id: int | None = None,
        no_clients: bool | None = None,
        total_currency: str | None = None,
        **kwargs,
    ) -> list[Project]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-projects."""
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'name': name,
                'name_or_code': name_or_code,
                'include_inputed_minutes': include_inputed_minutes,
                'updated_after': updated_after,
                'legal_entity_id': legal_entity_id,
                'no_clients': no_clients,
                'total_currency': total_currency,
            },
        )
        return _ProjectRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    @typing.overload
    async def get(self, *, project_id: int, **kwargs) -> Project: ...

    @typing.overload
    async def get(
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        name: str | None = None,
        name_or_code: str | None = None,
        include_inputed_minutes: bool | None = None,
        updated_after: datetime.date | None = None,
        legal_entity_id: int | None = None,
        no_clients: bool | None = None,
        total_currency: str | None = None,
        **kwargs,
    ) -> tuple[list[Project], _common.Meta]: ...

    async def get(  # noqa: PLR0913
        self,
        *,
        project_id: int | None = None,
        ids: typing.Sequence[int] | None = None,
        name: str | None = None,
        name_or_code: str | None = None,
        include_inputed_minutes: bool | None = None,
        updated_after: datetime.date | None = None,
        legal_entity_id: int | None = None,
        no_clients: bool | None = None,
        total_currency: str | None = None,
        **kwargs,
    ):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-projects-id."""
        if project_id is not None:
            return Project.model_validate(await self.api.get(self.endpoint, project_id, **kwargs))

        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'name': name,
                'name_or_code': name_or_code,
                'include_inputed_minutes': include_inputed_minutes,
                'updated_after': updated_after,
                'legal_entity_id': legal_entity_id,
                'no_clients': no_clients,
                'total_currency': total_currency,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _ProjectRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class ProjectTask(pydantic.BaseModel):
    id: int
    project_id: int
    subproject_id: int
    task_id: int
    follow_up: bool


class _ProjectTaskRoot(pydantic.RootModel):
    root: list[ProjectTask]


class ProjectTaskEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/project_management/project_tasks'

    async def all(  # noqa: PLR0913
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        task_ids: typing.Sequence[int] | None = None,
        completed: bool | None = None,
        overdue: bool | None = None,
        search: str | None = None,
        due_status: str | None = None,
        client_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> list[ProjectTask]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-project-tasks."""
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'project_ids[]': project_ids,
                'subproject_ids[]': subproject_ids,
                'task_ids[]': task_ids,
                'completed': completed,
                'overdue': overdue,
                'search': search,
                'due_status': due_status,
                'client_ids[]': client_ids,
            },
        )
        return _ProjectTaskRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    @typing.overload
    async def get(self, *, project_task_id: int, **kwargs) -> ProjectTask: ...

    @typing.overload
    async def get(
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        task_ids: typing.Sequence[int] | None = None,
        completed: bool | None = None,
        overdue: bool | None = None,
        search: str | None = None,
        due_status: str | None = None,
        client_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> tuple[list[ProjectTask], _common.Meta]: ...

    async def get(  # noqa: PLR0913
        self,
        *,
        project_task_id: int | None = None,
        ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        task_ids: typing.Sequence[int] | None = None,
        completed: bool | None = None,
        overdue: bool | None = None,
        search: str | None = None,
        due_status: str | None = None,
        client_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-project-tasks-id."""
        if project_task_id is not None:
            return ProjectTask.model_validate(await self.api.get(self.endpoint, project_task_id, **kwargs))
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'project_ids[]': project_ids,
                'subproject_ids[]': subproject_ids,
                'task_ids[]': task_ids,
                'completed': completed,
                'overdue': overdue,
                'search': search,
                'due_status': due_status,
                'client_ids[]': client_ids,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _ProjectTaskRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class ProjectWorker(pydantic.BaseModel):
    id: int
    project_id: int
    employee_id: int
    assigned: bool
    inputed_minutes: int | None
    labor_cost_cents: int | None
    spending_cost_cents: int | None


class _ProjectWorkerRoot(pydantic.RootModel):
    root: list[ProjectWorker]


class ProjectWorkerEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/project_management/project_workers'

    async def all(  # noqa: PLR0913
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        no_subproject: bool | None = None,
        employee_ids: typing.Sequence[int] | None = None,
        assigned: bool | None = None,
        project_active: bool | None = None,
        employee_name: str | None = None,
        include_inputed_minuted: bool | None = None,
        include_cost: bool | None = None,
        updated_after: datetime.date | None = None,
        include_labor_cost: bool | None = None,
        **kwargs,
    ) -> list[ProjectWorker]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-project-workers."""
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'project_ids[]': project_ids,
                'subproject_ids[]': subproject_ids,
                'no_subproject': no_subproject,
                'employee_ids[]': employee_ids,
                'assigned': assigned,
                'project_active': project_active,
                'employee_name': employee_name,
                'include_inputed_minutes': include_inputed_minuted,
                'include_cost': include_cost,
                'updated_after': updated_after,
                'include_labor_cost': include_labor_cost,
            },
        )
        return _ProjectWorkerRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    @typing.overload
    async def get(self, *, project_worker_id: int, **kwargs) -> ProjectWorker: ...

    @typing.overload
    async def get(
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        no_subproject: bool | None = None,
        employee_ids: typing.Sequence[int] | None = None,
        assigned: bool | None = None,
        project_active: bool | None = None,
        employee_name: str | None = None,
        include_inputed_minuted: bool | None = None,
        include_cost: bool | None = None,
        updated_after: datetime.date | None = None,
        include_labor_cost: bool | None = None,
        **kwargs,
    ) -> tuple[list[ProjectWorker], _common.Meta]: ...

    async def get(  # noqa: PLR0913
        self,
        *,
        project_worker_id: int | None = None,
        ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        no_subproject: bool | None = None,
        employee_ids: typing.Sequence[int] | None = None,
        assigned: bool | None = None,
        project_active: bool | None = None,
        employee_name: str | None = None,
        include_inputed_minuted: bool | None = None,
        include_cost: bool | None = None,
        updated_after: datetime.date | None = None,
        include_labor_cost: bool | None = None,
        **kwargs,
    ):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-project-workers-id."""
        if project_worker_id is not None:
            return ProjectWorker.model_validate(await self.api.get(self.endpoint, project_worker_id, **kwargs))
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'project_ids[]': project_ids,
                'subproject_ids[]': subproject_ids,
                'no_subproject': no_subproject,
                'employee_ids[]': employee_ids,
                'assigned': assigned,
                'project_active': project_active,
                'employee_name': employee_name,
                'include_inputed_minutes': include_inputed_minuted,
                'include_cost': include_cost,
                'updated_after': updated_after,
                'include_labor_cost': include_labor_cost,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _ProjectWorkerRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class Subproject(pydantic.BaseModel):
    id: int | None
    name: str
    project_id: int
    inputed_minutes: int | None
    labor_cost_cents: int | None


class _SubprojectRoot(pydantic.RootModel):
    root: list[Subproject]


class SubprojectEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/project_management/subprojects'

    async def all(  # noqa: PLR0913
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        name: str | None = None,
        include_no_subproject: bool | None = None,
        include_inputed_minuted: bool | None = None,
        include_cost: bool | None = None,
        updated_after: datetime.date | None = None,
        **kwargs,
    ) -> list[Subproject]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-subprojects."""
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'project_ids[]': project_ids,
                'name': name,
                'include_no_subproject': include_no_subproject,
                'include_inputed_minutes': include_inputed_minuted,
                'include_cost': include_cost,
                'updated_after': updated_after,
            },
        )
        return _SubprojectRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    @typing.overload
    async def get(self, *, subproject_id: int, **kwargs) -> Subproject: ...

    @typing.overload
    async def get(
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        name: str | None = None,
        include_no_subproject: bool | None = None,
        include_inputed_minuted: bool | None = None,
        include_cost: bool | None = None,
        updated_after: datetime.date | None = None,
        **kwargs,
    ) -> tuple[list[Subproject], _common.Meta]: ...

    async def get(  # noqa: PLR0913
        self,
        *,
        subproject_id: int | None = None,
        ids: typing.Sequence[int] | None = None,
        project_ids: typing.Sequence[int] | None = None,
        name: str | None = None,
        include_no_subproject: bool | None = None,
        include_inputed_minuted: bool | None = None,
        include_cost: bool | None = None,
        updated_after: datetime.date | None = None,
        **kwargs,
    ):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-subprojects-id."""
        if subproject_id is not None:
            return Subproject.model_validate(await self.api.get(self.endpoint, subproject_id, **kwargs))
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'project_ids[]': project_ids,
                'name': name,
                'include_no_subproject': include_no_subproject,
                'include_inputed_minutes': include_inputed_minuted,
                'include_cost': include_cost,
                'updated_after': updated_after,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _SubprojectRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class TimeRecord(pydantic.BaseModel):
    id: int
    project_worker_id: int
    attendance_shift_id: int
    subproject_id: int | None
    date: datetime.date | None
    imputed_minutes: int | None
    clock_in: datetime.datetime | None  # date will always be 2000-01-01, only use for .time()
    clock_out: datetime.datetime | None  # date will always be 2000-01-01, only use for .time()


class _TimeRecordRoot(pydantic.RootModel):
    root: list[TimeRecord]


class TimeRecordEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/project_management/time_records'

    async def all(  # noqa: PLR0913
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        project_worker_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        attendance_shift_ids: typing.Sequence[int] | None = None,
        month: int | None = None,
        year: int | None = None,
        updated_after: datetime.date | None = None,
        **kwargs,
    ) -> list[TimeRecord]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-time-records."""
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'project_worker_ids[]': project_worker_ids,
                'subproject_ids[]': subproject_ids,
                'attendance_shift_ids[]': attendance_shift_ids,
                'month': month,
                'year': year,
                'updated_after': updated_after,
            },
        )
        return _TimeRecordRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    @typing.overload
    async def get(self, *, time_record_id: int, **kwargs) -> TimeRecord: ...

    @typing.overload
    async def get(
        self,
        *,
        ids: typing.Sequence[int] | None = None,
        project_worker_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        attendance_shift_ids: typing.Sequence[int] | None = None,
        month: int | None = None,
        year: int | None = None,
        updated_after: datetime.date | None = None,
        **kwargs,
    ) -> tuple[list[TimeRecord], _common.Meta]: ...

    async def get(  # noqa: PLR0913
        self,
        *,
        time_record_id: int | None = None,
        ids: typing.Sequence[int] | None = None,
        project_worker_ids: typing.Sequence[int] | None = None,
        subproject_ids: typing.Sequence[int] | None = None,
        attendance_shift_ids: typing.Sequence[int] | None = None,
        month: int | None = None,
        year: int | None = None,
        updated_after: datetime.date | None = None,
        **kwargs,
    ):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-project-management-time-records-id."""
        if time_record_id is not None:
            return TimeRecord.model_validate(await self.api.get(self.endpoint, time_record_id, **kwargs))
        params = kwargs.get('params', {})
        params.update(
            {
                'ids[]': ids,
                'project_worker_ids[]': project_worker_ids,
                'subproject_ids[]': subproject_ids,
                'attendance_shift_ids[]': attendance_shift_ids,
                'month': month,
                'year': year,
                'updated_after': updated_after,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _TimeRecordRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])



import asyncio
import datetime

import click
import cloup
import tabulate

import factorialhr
from factorialhr._cli import account
from factorialhr._cli import common as common_cli


async def get_time_records(
    client: factorialhr.ApiClient,
    start: datetime.date | None,
    end: datetime.date,
) -> list[factorialhr.TimeRecord]:
    if start is not None:
        time_records = [
            x
            for xs in await asyncio.gather(
                *(
                    factorialhr.TimeRecordEndpoint(client).all(year=date.year, month=date.month)
                    for date in common_cli.get_months_in(start, end)
                ),
            )
            for x in xs
        ]
        return [
            time_record
            for time_record in time_records
            if time_record.date is not None and start <= time_record.date <= end
        ]
    return await factorialhr.TimeRecordEndpoint(client).all()


@cloup.group('projects', help='Project related cli methods.')
def project_entrypoint():
    pass


@cloup.command(help='Calculate the imputed minutes for projects')
@cloup.option_group(
    'Filter options',
    cloup.option('--name', type=str, help='Filter by project name'),
    cloup.option('--active', is_flag=True, help='Filter by active projects only'),
    *common_cli.upper_lower_bounds,
    cloup.option('--ignore-unassigned', is_flag=True, help='Ignore time records that are not assigned to any project'),
)
@common_cli.generic_options
@common_cli.to_async
async def imputed_minutes(  # noqa: PLR0913
    name: str | None,
    start: datetime.date | None,
    end: datetime.date,
    active: bool,  # noqa: FBT001
    less: bool,  # noqa: FBT001
    http_timeout: int,
    **_,
):
    auth, url = account.get_session()
    async with factorialhr.ApiClient(auth=auth, base_url=url, timeout=http_timeout) as client:
        time_records, projects, project_worker = await asyncio.gather(
            get_time_records(client, start, end),
            factorialhr.ProjectEndpoint(client).all(name=name),
            factorialhr.ProjectWorkerEndpoint(client).all(),
        )
    if not time_records:
        raise click.ClickException('No project time records found')
    if not projects:
        raise click.ClickException('No projects found')
    projects_by_id = {project.id: project for project in projects}
    project_workers_by_id = {project_worker.id: project_worker for project_worker in project_worker}

    # ensures that also projects that don't have imputed minutes are respected
    if active:
        projects_time = {project.name: 0 for project in projects if project.status == factorialhr.ProjectStatus.active}
    else:
        projects_time = {project.name: 0 for project in projects}
    for record in time_records:
        project_worker = project_workers_by_id[record.project_worker_id]
        if project_worker.project_id in projects_by_id:
            if record.imputed_minutes is None:
                msg = (
                    f'Time record contains no imputed minutes. This might happen if a time record has no end time yet '
                    f'(e.g. is someone is currently working). Record: "{record}"'
                )
                raise click.ClickException(msg)
            projects_time[projects_by_id[project_worker.project_id].name] += record.imputed_minutes

    table_to_print = tabulate.tabulate(list(projects_time.items()), headers=['Projectname', 'Imputed minutes'])
    if less:
        click.echo_via_pager(table_to_print)
    else:
        click.echo(table_to_print)


@cloup.command(help='Calculate the minutes not assigned to any project')
@cloup.option_group(
    'Filter options',
    *common_cli.upper_lower_bounds,
)
@common_cli.generic_options
@common_cli.to_async
async def unrelated_minutes(
    start: datetime.date | None,
    end: datetime.date,
    http_timeout: int,
    **_,
):
    auth, url = account.get_session()
    async with factorialhr.ApiClient(auth=auth, base_url=url, timeout=http_timeout) as client:
        shifts, time_records = await asyncio.gather(
            # get raw data as its not worth parsing the data because only integer are used
            factorialhr.ShiftEndpoint(client).all_raw(params={'start_on': start, 'end_on': end}),
            get_time_records(client, start, end),
        )
    if not shifts:
        raise click.ClickException('No shifts found')
    if not time_records:
        raise click.ClickException('No project time records found')
    records_by_shift_id = {record.attendance_shift_id: record for record in time_records}
    click.echo(f'{sum(s["minutes"] for s in shifts if s["id"] not in records_by_shift_id)} minutes')


project_entrypoint.add_command(imputed_minutes)
project_entrypoint.add_command(unrelated_minutes)



import enum
import typing

import pydantic


class Meta(pydantic.BaseModel):
    has_next_page: bool
    has_previous_page: bool
    start_cursor: str | None = pydantic.Field(default=None)
    end_cursor: str | None = pydantic.Field(default=None)
    limit: int | None  # apparently this is can be None sometimes, e.g. when specifying employee_ids[] in shift request
    total: int


class LocationType(enum.StrEnum):
    office = 'office'
    business_trip = 'business_trip'
    work_from_home = 'work_from_home'


class TimeUnit(enum.StrEnum):
    minute = 'minute'
    half_day = 'half_day'
    none = 'none'


def build_params(**kwargs) -> dict[str, typing.Any]:
    """Remove None values from kwargs."""
    return {key: value for key, value in kwargs.items() if value is not None}



import datetime
import enum
import typing

import pydantic

from factorialhr import _common
from factorialhr._client import Endpoint


class BreakConfiguration(pydantic.BaseModel):
    id: int
    attendance_employees_setting_id: int
    time_settings_break_configuration_id: int
    enabled: bool
    name: str | None
    paid: bool | None


class _BreakConfigurationRoot(pydantic.RootModel):
    root: list[BreakConfiguration]


class BreakConfigurationEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/attendance/break_configurations'

    async def all(self, **kwargs) -> list[BreakConfiguration]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-break-configurations."""
        return _BreakConfigurationRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, break_configuration_id: int, **kwargs) -> BreakConfiguration: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[BreakConfiguration], _common.Meta]: ...

    async def get(self, *, break_configuration_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-break-configurations-id."""
        result = await self.api.get(self.endpoint, break_configuration_id, **kwargs)
        return (
            BreakConfiguration.model_validate(result)
            if break_configuration_id is not None
            else (
                _BreakConfigurationRoot.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )


class EditTimesheetRequestRequestType(enum.StrEnum):
    create_shift = 'create_shift'
    delete_shift = 'delete_shift'
    update_shift = 'update_shift'


class EditTimesheetRequest(pydantic.BaseModel):
    id: int
    approved: bool | None
    request_type: EditTimesheetRequestRequestType
    employee_id: int
    workable: bool | None
    clock_in: datetime.datetime | None
    clock_out: datetime.datetime | None
    location_type: _common.LocationType | None
    reason: str | None
    attendance_shift_id: int | None
    time_settings_break_configuration_id: int | None
    observations: str | None
    date: datetime.date | None
    reference_date: str | None


class _EditTimesheetRequestRoot(pydantic.RootModel):
    root: list[EditTimesheetRequest]


class EditTimesheetRequestEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/attendance/edit_timesheet_requests'

    async def all(self, **kwargs) -> list[EditTimesheetRequest]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-edit-timesheet-requests."""
        return _EditTimesheetRequestRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, break_configuration_id: int, **kwargs) -> EditTimesheetRequest: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[EditTimesheetRequest], _common.Meta]: ...
    async def get(self, *, edit_time_sheet_request_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-edit-timesheet-requests-id."""
        result = await self.api.get(self.endpoint, edit_time_sheet_request_id, **kwargs)
        return (
            EditTimesheetRequest.model_validate(result)
            if edit_time_sheet_request_id is not None
            else (
                _EditTimesheetRequestRoot.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )


class EstimatedTimeSource(enum.StrEnum):
    shift_management = 'shift_management'
    work_schedule = 'work_schedule'
    contract_hours = 'contract_hours'
    none = 'none'


class EstimatedTime(pydantic.BaseModel):
    date: datetime.date
    company_id: int
    employee_id: int
    expected_minutes: int
    regular_minutes: int
    overtime_minutes: int
    breaks: list[typing.Any]  # TODO: what is the break model?
    time_unit: _common.TimeUnit
    estimated_half_days: int
    shifts: list[typing.Any]  # TODO: what is the shifts model?
    source: str
    minutes: int


class _EstimatedTimeRoot(pydantic.RootModel):
    root: list[EstimatedTime]


class EstimatedTimeEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/attendance/estimated_times'

    async def all(
        self,
        *,
        start_on: datetime.date,
        end_on: datetime.date,
        employee_ids: typing.Sequence[int],
        **kwargs,
    ) -> list[EstimatedTime]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-estimated-times."""
        params = kwargs.get('params', {})
        params.update({'start_on': start_on, 'end_on': end_on, 'employee_ids[]': employee_ids})
        return _EstimatedTimeRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    async def get(
        self,
        *,
        start_on: datetime.date,
        end_on: datetime.date,
        employee_ids: typing.Sequence[int],
        **kwargs,
    ) -> tuple[list[EstimatedTime], _common.Meta]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-estimated-times."""
        params = kwargs.get('params', {})
        params.update({'start_on': start_on, 'end_on': end_on, 'employee_ids[]': employee_ids})
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _EstimatedTimeRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class OpenShiftStatus(enum.StrEnum):
    opened = 'opened'
    closed = 'closed'
    created = 'created'


class OpenShift(pydantic.BaseModel):
    id: int
    employee_id: int
    date: str
    clock_in: datetime.datetime
    clock_out: datetime.datetime | None
    status: OpenShiftStatus
    workable: bool
    automatic_clock_in: bool
    location_type: str | None
    workplace_id: int | None
    time_settings_break_configuration_id: int | None


class _OpenShiftRoot(pydantic.RootModel):
    root: list[OpenShift]


class OpenShiftEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/attendance/open_shifts'

    async def all(self, employee_ids: typing.Sequence[int] | None = None, **kwargs) -> list[OpenShift]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-open-shifts."""
        params = kwargs.get('params', {})
        params.update({'employee_ids[]': employee_ids} if employee_ids is not None else {})
        return _OpenShiftRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    async def get(
        self,
        employee_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> tuple[list[OpenShift], _common.Meta]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-open-shifts."""
        params = kwargs.get('params', {})
        params.update({'employee_ids[]': employee_ids} if employee_ids is not None else {})
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _OpenShiftRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class OverTimeRequestStatus(enum.StrEnum):
    pending = 'pending'
    approved = 'approved'
    rejected = 'rejected'
    none = 'none'


class OverTimeRequest(pydantic.BaseModel):
    id: int
    employee_id: int
    approver_id: int
    author_id: int
    status: OverTimeRequestStatus
    description: str | None
    reason: str | None
    date: datetime.date
    hours_amount_in_cents: int
    created_at: datetime.datetime | None
    approver: bool
    approver_full_name: str | None
    is_editable: bool


class _OverTimeRequestRoot(pydantic.RootModel):
    root: list[OverTimeRequest]


class OverTimeRequestEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/attendance/overtime_requests'

    async def all(self, **kwargs) -> list[OverTimeRequest]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-overtime-requests."""
        return _OverTimeRequestRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, answer_id: int, **kwargs) -> OverTimeRequest: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[OverTimeRequest], _common.Meta]: ...

    async def get(self, *, over_time_request_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-overtime-requests-id."""
        result = await self.api.get(self.endpoint, over_time_request_id, **kwargs)
        return (
            OverTimeRequest.model_validate(result)
            if over_time_request_id is not None
            else (
                _OverTimeRequestRoot.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )


class ShiftHalfDay(enum.StrEnum):
    beginning_of_day = 'beginning_of_day'
    end_of_day = 'end_of_day'


class Shift(pydantic.BaseModel):
    id: int
    employee_id: int
    date: datetime.date
    reference_date: datetime.date
    clock_in: datetime.time | None
    clock_out: datetime.time | None
    in_source: str | None
    out_source: str | None
    observations: str | None
    location_type: _common.LocationType | None
    half_day: ShiftHalfDay | None
    in_location_latitude: float | None
    in_location_longitude: float | None
    in_location_accuracy: int | None
    out_location_latitude: float | None
    out_location_longitude: float | None
    out_location_accuracy: float | None
    workable: bool | None
    created_at: datetime.datetime
    workplace_id: int | None
    time_settings_break_configuration_id: int | None
    company_id: int
    updated_at: str
    minutes: int
    clock_in_with_seconds: datetime.time | None


class _ShiftRoot(pydantic.RootModel):
    root: list[Shift]


class ShiftEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/attendance/shifts'

    async def all(  # noqa: PLR0913
        self,
        *,
        employee_ids: typing.Sequence[int] | None = None,
        start_on: datetime.date | None = None,
        end_on: datetime.date | None = None,
        ids: typing.Sequence[int] | None = None,
        half_day: bool | None = None,
        workable: bool | None = None,
        latest_shift: bool | None = None,
        sort_created_at_asc: bool | None = None,
        breaks_with_time_configuration: bool | None = None,
        last_working_shift: bool | None = None,
        **kwargs,
    ) -> list[Shift]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-shifts."""
        params = kwargs.get('params', {})
        params.update(
            {
                'employee_ids[]': employee_ids,
                'start_on': start_on,
                'end_on': end_on,
                'ids[]': ids,
                'half_day': half_day,
                'workable': workable,
                'latest_shift': latest_shift,
                'sort_created_at_asc': sort_created_at_asc,
                'breaks_with_time_configuration': breaks_with_time_configuration,
                'last_working_shift': last_working_shift,
            },
        )
        return _ShiftRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    @typing.overload
    async def get(self, *, shift_id: int, **kwargs) -> Shift: ...

    @typing.overload
    async def get(
        self,
        *,
        employee_ids: typing.Sequence[int] | None = None,
        start_on: datetime.date | None = None,
        end_on: datetime.date | None = None,
        ids: typing.Sequence[int] | None = None,
        half_day: bool | None = None,
        workable: bool | None = None,
        latest_shift: bool | None = None,
        sort_created_at_asc: bool | None = None,
        breaks_with_time_configuration: bool | None = None,
        last_working_shift: bool | None = None,
        **kwargs,
    ) -> tuple[list[Shift], _common.Meta]: ...
    async def get(  # noqa: PLR0913
        self,
        *,
        shift_id: int | None = None,
        employee_ids: typing.Sequence[int] | None = None,
        start_on: datetime.date | None = None,
        end_on: datetime.date | None = None,
        ids: typing.Sequence[int] | None = None,
        half_day: bool | None = None,
        workable: bool | None = None,
        latest_shift: bool | None = None,
        sort_created_at_asc: bool | None = None,
        breaks_with_time_configuration: bool | None = None,
        last_working_shift: bool | None = None,
        **kwargs,
    ):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-shifts-id."""
        if shift_id is not None:
            return Shift.model_validate(await self.api.get(self.endpoint, shift_id, **kwargs))
        params = kwargs.get('params', {})
        params.update(
            {
                'employee_ids[]': employee_ids,
                'start_on': start_on,
                'end_on': end_on,
                'ids[]': ids,
                'half_day': half_day,
                'workable': workable,
                'latest_shift': latest_shift,
                'sort_created_at_asc': sort_created_at_asc,
                'breaks_with_time_configuration': breaks_with_time_configuration,
                'last_working_shift': last_working_shift,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _ShiftRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])


class WorkedTimeDayType(enum.StrEnum):
    saturday = 'saturday'
    sunday = 'sunday'
    bank_holiday = 'bank_holiday'
    workday = 'workday'


class WorkedTime(pydantic.BaseModel):
    employee_id: int
    date: datetime.date
    company_id: int
    tracked_minutes: int
    multiplied_minutes: int
    pending_minutes: int
    minutes: int
    time_unit: _common.TimeUnit
    worked_time_blocks: list[typing.Any]
    day_type: WorkedTimeDayType
    id: str


class _WorkedTimeRoot(pydantic.RootModel):
    root: list[WorkedTime]


class WorkedTimeEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/attendance/worked_times'

    async def all(
        self,
        *,
        include_time_range_category: bool,
        include_non_attendable_employees: bool,
        start_on: datetime.date | None = None,
        end_on: datetime.date | None = None,
        employee_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> list[WorkedTime]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-worked-times."""
        params = kwargs.get('params', {})
        params.update(
            {
                'include_time_range_category': include_time_range_category,
                'include_non_attendable_employees': include_non_attendable_employees,
                'start_on': start_on,
                'end_on': end_on,
                'employee_ids[]': employee_ids,
            },
        )
        return _WorkedTimeRoot.model_validate(
            await self.api.get_all(self.endpoint, params=_common.build_params(**params), **kwargs),
        ).root

    async def get(
        self,
        *,
        include_time_range_category: bool,
        include_non_attendable_employees: bool,
        start_on: datetime.date | None = None,
        end_on: datetime.date | None = None,
        employee_ids: typing.Sequence[int] | None = None,
        **kwargs,
    ) -> tuple[list[WorkedTime], _common.Meta]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-attendance-worked-times."""
        params = kwargs.get('params', {})
        params.update(
            {
                'include_time_range_category': include_time_range_category,
                'include_non_attendable_employees': include_non_attendable_employees,
                'start_on': start_on,
                'end_on': end_on,
                'employee_ids[]': employee_ids,
            },
        )
        result = await self.api.get(self.endpoint, params=_common.build_params(**params), **kwargs)
        return _WorkedTimeRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta'])



import datetime
import enum
import typing

import pydantic

from factorialhr import _common
from factorialhr._client import Endpoint


class AnswerOriginalQuestionType(enum.StrEnum):
    text = 'text'
    long_text = 'long_text'
    single_choice = 'single_choice'
    multiple_choice = 'multiple_choice'
    file = 'file'


class Answer(pydantic.BaseModel):
    id: int
    ats_question_id: int | None
    ats_application_id: int
    original_question_label: str
    value: str | None
    original_question_type: AnswerOriginalQuestionType
    created_at: datetime.datetime
    updated_at: datetime.datetime


class _AnswerRootModel(pydantic.RootModel):
    root: list[Answer]


class AnswerEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/ats/answers'

    async def all(self, **kwargs) -> list[Answer]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-answers."""
        return _AnswerRootModel.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, answer_id: int, **kwargs) -> Answer: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[Answer], _common.Meta]: ...

    async def get(self, *, answer_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-answers-id."""
        result = await self.api.get(self.endpoint, answer_id, **kwargs)
        return (
            Answer.model_validate(result)
            if answer_id is not None
            else (
                _AnswerRootModel.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )


class Application(pydantic.BaseModel):
    id: int
    company_id: int
    ats_job_posting_id: int
    ats_candidate_id: int
    employee_id: int | None
    phone: str | None
    qualified: bool | None
    ats_application_phase_id: int | None
    created_at: datetime.datetime
    cover_letter: str | None
    ats_conversation_id: int | None
    medium: str | None
    rating_average: int | None
    ats_rejection_reason_id: int | None
    source_id: int | None


class _ApplicationRoot(pydantic.RootModel):
    root: list[Application]


class ApplicationEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/ats/applications'

    async def all(self, **kwargs) -> list[Application]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-applications."""
        return _ApplicationRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, application_id: int, **kwargs) -> Application: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[Application], _common.Meta]: ...

    async def get(self, *, application_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-applications-id."""
        result = await self.api.get(self.endpoint, application_id, **kwargs)
        return (
            Application.model_validate(result)
            if application_id is not None
            else (
                _ApplicationRoot.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )


class ApplicationPhasePhaseType(enum.StrEnum):
    initial = 'initial'
    normal = 'normal'
    hired = 'hired'
    screening = 'screening'
    interview = 'interview'
    assessment = 'assessment'
    offer = 'offer'


class ApplicationPhase(pydantic.BaseModel):
    id: int
    ats_job_posting_id: int
    name: str
    position: int
    editable: bool
    phase_type: ApplicationPhasePhaseType
    applications_count: int | None
    active_applications_count: int | None
    ats_hiring_stage_id: int | None


class _ApplicationPhaseRoot(pydantic.RootModel):
    root: list[ApplicationPhase]


class ApplicationPhaseEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/ats/application_phases'

    async def all(self, **kwargs) -> list[ApplicationPhase]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-application-phases."""
        return _ApplicationPhaseRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, application_phase: int, **kwargs) -> ApplicationPhase: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[ApplicationPhase], _common.Meta]: ...

    async def get(self, *, application_phase_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-application-phases-id."""
        result = await self.api.get(self.endpoint, application_phase_id, **kwargs)
        return (
            ApplicationPhase.model_validate(result)
            if application_phase_id is not None
            else (
                _ApplicationPhaseRoot.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )


class CandidateGender(enum.StrEnum):
    female = 'female'
    male = 'male'
    unanswered = 'unanswered'
    other = 'other'


class Candidate(pydantic.BaseModel):
    id: int
    company_id: int | None
    first_name: str
    last_name: str
    full_name: str
    email: str | None
    talent_pool: bool
    phone_number: int | None
    created_at: datetime.datetime
    updated_at: datetime.datetime
    consent_given_at: datetime.datetime | None
    inactive_since: datetime.datetime | None
    ats_job_posting_ids: list[int] | None
    personal_url: str | None
    consent_expiration_date: datetime.datetime | None
    consent_to_talent_pool: bool | None
    medium: str | None
    source_id: int | None
    gender: CandidateGender | None
    score: int | None


class _CandidateRoot(pydantic.RootModel):
    root: list[Candidate]


class CandidateEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/ats/candidates'

    async def all(self, **kwargs) -> list[Candidate]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-candidates."""
        return _CandidateRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, candidate_id: int, **kwargs) -> Candidate: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[Candidate], _common.Meta]: ...

    async def get(self, *, candidate_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-candidates-id."""
        result = await self.api.get(self.endpoint, candidate_id, **kwargs)
        return (
            Candidate.model_validate(result)
            if candidate_id is not None
            else (
                _CandidateRoot.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )


class CandidateSourceCategory(enum.StrEnum):
    job_board = 'job_board'
    referral = 'referral'
    social_media = 'social_media '
    agency_or_external_recruiter = 'agency_or_external_recruiter '
    event = 'event'
    internal = 'internal'
    manually_added = 'manually_added'
    organic = 'organic'


class CandidateSource(pydantic.BaseModel):
    id: int
    company_id: int
    category: CandidateSourceCategory
    name: str
    label: str


class _CandidateSourceRoot(pydantic.RootModel):
    root: list[CandidateSource]


class CandidateSourceEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/ats/candidate_sources'

    async def all(self, **kwargs) -> list[CandidateSource]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-candidate-sources."""
        return _CandidateSourceRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, candidate_source_id: int, **kwargs) -> CandidateSource: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[CandidateSource], _common.Meta]: ...

    async def get(self, *, candidate_source_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-candidate-sources-id."""
        result = await self.api.get(self.endpoint, candidate_source_id, **kwargs)
        return (
            CandidateSource.model_validate(result)
            if candidate_source_id is not None
            else (
                _CandidateSourceRoot.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )


class EvaluationFormQuestion(pydantic.BaseModel):
    id: str  # should be int but the API returns a string (could be an uuid)
    text: str
    description: str


class EvaluationForm(pydantic.BaseModel):
    id: int
    company_id: int
    ats_job_posting_id: int | None
    name: str
    based_on_id: int | None
    questions: list[EvaluationFormQuestion]
    created_at: datetime.datetime
    updated_at: datetime.datetime


class _EvaluationFormRoot(pydantic.RootModel):
    root: list[EvaluationForm]


class EvaluationFormEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/ats/evaluation_forms'

    async def all(self, **kwargs) -> list[EvaluationForm]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-evaluation-forms."""
        return _EvaluationFormRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, evaluation_form_id: int, **kwargs) -> EvaluationForm: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[EvaluationForm], _common.Meta]: ...

    async def get(self, *, evaluation_form_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-evaluation-forms-id."""
        result = await self.api.get(self.endpoint, evaluation_form_id, **kwargs)
        return (
            EvaluationForm.model_validate(result)
            if evaluation_form_id is not None
            else (
                _EvaluationFormRoot.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )


class FeedbackEvaluationFormAnswer(pydantic.BaseModel):
    id: int
    score: int
    note: str


class Feedback(pydantic.BaseModel):
    id: int
    rating: int | None
    description: str | None
    ats_application_id: int | None
    ats_application_phase_id: int | None
    created_at: datetime.datetime
    ats_candidate_id: int
    ats_evaluation_forms_id: int | None
    evaluation_form_answers: list[FeedbackEvaluationFormAnswer] | None


class _FeedbackRoot(pydantic.RootModel):
    root: list[Feedback]


class FeedbackEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/ats/feedbacks'

    async def all(self, **kwargs) -> list[Feedback]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-feedbacks."""
        return _FeedbackRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, feedback_id: int, **kwargs) -> Feedback: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[Feedback], _common.Meta]: ...

    async def get(self, *, feedback_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-feedbacks-id."""
        result = await self.api.get(self.endpoint, feedback_id, **kwargs)
        return (
            Feedback.model_validate(result)
            if feedback_id is not None
            else (_FeedbackRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta']))
        )


class HiringStateName(enum.StrEnum):
    new = 'new'
    screening = 'screening'
    interview = 'interview'
    assessment = 'assessment'
    offer = 'offer'
    hired = 'hired'


class HiringStage(pydantic.BaseModel):
    id: int
    name: HiringStateName
    company_id: int
    position: int


class _HiringStageRoot(pydantic.RootModel):
    root: list[HiringStage]


class HiringStageEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/ats/hiring_stages'

    async def all(self, **kwargs) -> list[HiringStage]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-hiring-stages."""
        return _HiringStageRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, hiring_stage_id: int, **kwargs) -> HiringStage: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[HiringStage], _common.Meta]: ...

    async def get(self, *, hiring_stage_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-hiring-stages-id."""
        result = await self.api.get(self.endpoint, hiring_stage_id, **kwargs)
        return (
            HiringStage.model_validate(result)
            if hiring_stage_id is not None
            else (
                _HiringStageRoot.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )


class JobPostingStatus(enum.StrEnum):
    draft = 'draft'
    published = 'published'
    unlisted = 'unlisted'
    archived = 'archived'
    cancelled = 'cancelled'


class JobPostingSalaryPeriod(enum.StrEnum):
    annual = 'annual'
    monthly = 'monthly'
    daily = 'daily'


class JobPosting(pydantic.BaseModel):
    id: int
    company_id: int
    ats_company_id: str
    title: str
    description: str | None
    contract_type: str | None
    workplace_type: str | None
    remote: bool
    status: JobPostingStatus
    schedule_type: str | None
    team_id: str | None
    location_id: str | None
    legal_entity_id: str | None
    salary_format: str | None
    salary_from_amount_in_cents: str | None
    salary_to_amount_in_cents: str | None
    hide_salary: str | None
    cv_requirement: str
    cover_letter_requirement: str
    phone_requirement: str
    photo_requirement: str
    preview_token: str | None
    url: str | None
    salary_period: JobPostingSalaryPeriod
    published_at: datetime.datetime | None
    created_at: datetime.datetime


class _JobPostingRootModel(pydantic.RootModel):
    root: list[JobPosting]


class JobPostingEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/ats/job_postings'

    async def all(self, **kwargs) -> list[JobPosting]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-job-postings."""
        return _JobPostingRootModel.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, job_posting_id: int, **kwargs) -> JobPosting: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[JobPosting], _common.Meta]: ...

    async def get(self, *, job_posting_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-job-postings-id."""
        result = await self.api.get(self.endpoint, job_posting_id, **kwargs)
        return (
            JobPosting.model_validate(result)
            if job_posting_id is not None
            else (
                _JobPostingRootModel.model_validate(result['data']).root,
                _common.Meta.model_validate(result['meta']),
            )
        )


class Message(pydantic.BaseModel):
    id: int
    content: str
    ats_conversation_id: int
    sent_by_id: int
    sent_by_type: str
    created_at: datetime.datetime
    attachments: list[typing.Any]  # TODO: what is the attachment model?
    topic: str
    delayed_until: datetime.datetime | None
    sent_at: datetime.datetime | None
    delivered_at: datetime.datetime | None
    opened_at: datetime.datetime | None
    last_error_at: datetime.datetime | None


class _MessageRoot(pydantic.RootModel):
    root: list[Message]


class MessageEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/ats/messages'

    async def all(self, **kwargs) -> list[Message]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-messages."""
        return _MessageRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, evaluation_form_id: int, **kwargs) -> Message: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[Message], _common.Meta]: ...

    async def get(self, *, evaluation_form_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-messages-id."""
        result = await self.api.get(self.endpoint, evaluation_form_id, **kwargs)
        return (
            Message.model_validate(result)
            if evaluation_form_id is not None
            else (_MessageRoot.model_validate(result['data']).root, _common.Meta.model_validate(result['meta']))
        )


class RejectReasonDecisionMaker(enum.StrEnum):
    company = 'company'
    candidate = 'candidate'


class RejectReason(pydantic.BaseModel):
    id: int
    company_id: int
    decision_maker: RejectReasonDecisionMaker
    reason: str
    created_at: datetime.datetime
    updated_at: datetime.datetime


class _RejectReasonRoot(pydantic.RootModel):
    root: list[RejectReason]


class RejectReasonEndpoint(Endpoint):
    endpoint = '/2025-01-01/resources/ats/rejection_reasons'

    async def all(self, **kwargs) -> list[RejectReason]:
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-rejection-reasons."""
        return _RejectReasonRoot.model_validate(await self.api.get_all(self.endpoint, **kwargs)).root

    @typing.overload
    async def get(self, *, reject_reason_id: int, **kwargs) -> RejectReason: ...

    @typing.overload
    async def get(self, **kwargs) -> tuple[list[RejectReason], _common.Meta]: ...

    async def get(self, *, reject_reason_id: int | None = None, **kwargs):
        """Implement https://apidoc.factorialhr.com/reference/get_api-2025-01-01-resources-ats-rejection-reasons-id."""
        result = await self.api.get(self.endpoint, reject_reason_id, **kwargs)
        return (
            RejectReason.model_validate(result)
            if reject_reason_id is not None
            else (
                _RejectReasonRoot.model_validate(result).root,
                _common.Meta.model_validate(result['meta']),
            )
        )



"""Implementation of an HTTP server using aiohttp for handling OAuth2 authorization code flow."""

import asyncio
import contextlib
import ssl
import uuid
from collections.abc import Awaitable, Callable

from aiohttp import web


class HttpServer:
    """An HTTP server for handling OAuth2 authorization code flow."""

    def __init__(
        self,
        host: str,
        port: int,
        handler: Callable[[web.Request], Awaitable[web.StreamResponse]],
        ssl_context: ssl.SSLContext | None = None,
    ):
        self.host = host
        self.port: int = port
        self.ssl_context = ssl_context
        app = web.Application()
        app.add_routes([web.get('/authorize/authorization_code', handler)])
        self.runner = web.AppRunner(app)
        self._server_task: asyncio.Task | None = None
        self._server_started = asyncio.Event()
        self._server_close_request = asyncio.Event()
        self.state_secret = uuid.uuid4().hex

    async def __aenter__(self) -> 'HttpServer':
        await self.start()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.close()

    async def _start_http_server(self):
        await self.runner.setup()
        self._site = web.TCPSite(self.runner, self.host, self.port, ssl_context=self.ssl_context)
        await self._site.start()
        self._server_started.set()
        try:
            await self._server_close_request.wait()
        finally:
            await self.runner.cleanup()

    async def start(self):
        """Start the server."""
        self._server_task = asyncio.create_task(self._start_http_server())
        await self._server_started.wait()

    async def close(self):
        """Close the server."""
        self._server_started.clear()
        self._server_close_request.set()
        if self._server_task:
            self._server_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._server_task
